// gpusb_test, quick and dirty test application to control Shoestring's GPUSB.
// Functionally equivalent to the test software that Shoeastring distribute for windows.
// Basic routines will of course be useful to add what is needed to control the GPUSB smart box from within homebrew softwares.
// It may be also useful for Mac programmers as long as libusb1.0 can be installed. I'm currently not certain about this.
//
// author: Giampiero Spezzano gspezzano@gmail.com
//
// free to distribute, update, upgrade, include in other projects when full credit to author is granted :-)
//
// requires libusb1.0 (also dev) and lesstif (also dev). Dev packages are of couse only needed for compiling.
//
// please compile with:
// gcc gpusb_test.c -lXm -lXt -lXmu -lXp -lXext -lX11 -lSM -lICE -lm -lc -lusb-1.0 -lpthread -lrt -o gpusb_test

#include <libusb-1.0/libusb.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include <Xm/Xm.h>
#include <Xm/Protocols.h>
#include <Xm/Form.h>
#include <Xm/Separator.h>
#include <Xm/RowColumn.h> 
#include <Xm/Label.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/PushB.h>
#include <Xm/ArrowBG.h>
#include <Xm/ToggleB.h>
#include <Xm/SpinB.h>
#include <Xm/List.h>
#include <Xm/FileSB.h>
#include <Xm/CascadeB.h>
#include <Xm/MessageB.h>
#include <Xm/SelectioB.h>

#define VENDOR_ID   0x134A
#define PRODUCT_ID  0x9020

// HID Class-Specific Requests values. See section 7.2 of the HID specifications 

#define HID_GET_REPORT                0x01
#define HID_GET_IDLE                  0x02
#define HID_GET_PROTOCOL              0x03
#define HID_SET_REPORT                0x09
#define HID_SET_IDLE                  0x0A
#define HID_SET_PROTOCOL              0x0B
#define HID_REPORT_TYPE_INPUT         0x01
#define HID_REPORT_TYPE_OUTPUT        0x02
#define HID_REPORT_TYPE_FEATURE       0x03
#define CTRL_IN     LIBUSB_ENDPOINT_IN
#define CTRL_OUT    LIBUSB_ENDPOINT_OUT

#define ENDPOINT_INT_IN 0x81
#define ENDPOINT_INT_OUT 0x02

#define TIMEOUT    1000
#define PACKET_LEN 1 

/* Slew directions */
#define	DECP		1
#define	DECM		2
#define	RAM 		4
#define	RAP 		8

/* CmdSlew directions */
#define	CMDNORTH	0
#define	CMDEAST		1
#define	CMDWEST		2
#define	CMDSOUTH	3

/* Command Buttons */
#define CMD_BUTTONS 	3
#define CMD_CONNECT 	0
#define CMD_LEDCOL  	1
#define CMD_LEDENA  	2

XtAppContext context;

/* Widgets for the panel */
Widget toplevel;
Widget form;
Widget button[4];
Widget toggleS[2];
Widget command[CMD_BUTTONS];
Widget message_area;

/* Labels for the guide buttons top to bottom then left to right */
static int btn_name[4] = {
    XmARROW_UP,
    XmARROW_LEFT,
    XmARROW_RIGHT,
    XmARROW_DOWN
};

static int btn_pos[4][2] = {
    { 84,  34},
    { 14,  98},
    {154,  98},
    { 84, 158}
};

/* Simple labels for the command area top to bottom, left to right */
static char cmd_name[CMD_BUTTONS][15] = {
    "Connect",
    "Led Color",
    "Led Enable"
};

static int cmd_pos[CMD_BUTTONS][4] = {
    {81, 101, 71, 56},
    {14, 184, 71, 28},
    {154, 184, 71, 28}
};

static struct libusb_device_handle *hDevice = NULL;
static int device_link = 0;
static char strmsg[256];

union {
	unsigned char whole_byte;
	struct
	{
		 unsigned int FRAM:1;
		 unsigned int FRAP:1;
		 unsigned int FDECM:1;
		 unsigned int FDECP:1;
		 unsigned int FLEDColor:1;
		 unsigned int FLEDEnable:1;
		 unsigned int FRes1:1;
		 unsigned int FRes2:1;
	};
} devmessage;

static unsigned char databuffer[PACKET_LEN];

/* End of global definitions */

/* Update message area of user interface */
void setMessage(char *setstr)
{
	Arg al[15];
	int ac;
	
	XmTextInsert(message_area, XmTextGetLastPosition(message_area), setstr);
	ac=0;
	XtSetArg(al[ac],XmNvalue, 0); ac++;
	XtSetValues(message_area, al, ac);
} 

/* Basic usb functions */
int find_device(int vendorid, int productid, libusb_device_handle **handle) 
{
	libusb_device **list;
	libusb_device *found = NULL;
	ssize_t cnt = libusb_get_device_list(NULL, &list);
	ssize_t i = 0;
	struct libusb_device_descriptor dev_desc;
	int retcode = 0;
	
	if (cnt >= 0)
	{
		for (i = 0; i < cnt; i++) {
			libusb_device *device = list[i];
		    if (libusb_get_device_descriptor(device, &dev_desc) == 0) 
		    {
				if (dev_desc.idVendor == vendorid && dev_desc.idProduct == productid) 
				{
					found = device;
					break;
				}
		    }
		}

		if (found) 
		{
			if (libusb_open(found, handle) == 0)
			{
				retcode = 1;
			}
			else
			{
				setMessage("Error: Device found, but did not open!\n");
				found = NULL;
			}
		}
		else
		{
			setMessage("Error: Device NOT found!\n");
			found = NULL;
		}
	}
	libusb_free_device_list(list, 1);
	return retcode;
}

int OpenDevice(int vendorid, int productid) 
{
	int retcode = 1;

	retcode = (libusb_init(NULL) == 0);
	if(retcode)
	{
		// Level 0: no messages ever printed by the library (default)
		// Level 1: error messages are printed to stderr
		// Level 2: warning and error messages are printed to stderr
		// Level 3: informational messages are printed to stdout, warning and error messages are printed to stderr
		libusb_set_debug(NULL,0);
		if (find_device(vendorid, productid, &hDevice))
		{

			if (libusb_kernel_driver_active(hDevice, 0))
			{
				libusb_detach_kernel_driver(hDevice, 0);
			}
			if (libusb_set_configuration(hDevice,1) == 0) 
			{
				if (libusb_claim_interface(hDevice, 0) != 0) 
				{
					setMessage("Error: Could not claim interface:\n");
					retcode = 0;
				}
			}
			else
			{
				setMessage("Error: Could not set device configuration:\n");
				retcode = 0;
			}
		}
		else
		{
			retcode = 0;
		}
	}
	else
	{
		setMessage("Error: Could not initialise libusb\n");
		retcode = 0;
	}
	return retcode;
}

int CloseDevice() 
{
    libusb_release_interface(hDevice, 0); 
    //libusb_reset_device(hDevice);
	libusb_close(hDevice);
	libusb_exit(NULL);
	return 1;
}

int ToDevice( unsigned char *buf, int size) 
{
	int rcode, transferred=0;
	rcode = libusb_bulk_transfer(hDevice, ENDPOINT_INT_OUT, buf, size, &transferred, TIMEOUT); 
	return (rcode==0);
}

int FromDevice( unsigned char *buf, int size) 
{
	int rcode, transferred=0;
	//Must be read twice to get the right value, can't get why but it's recurring when programming libusb1.0
	rcode = libusb_interrupt_transfer(hDevice, ENDPOINT_INT_IN, buf, size, &transferred, TIMEOUT); 
	rcode = libusb_interrupt_transfer(hDevice, ENDPOINT_INT_IN, buf, size, &transferred, TIMEOUT); 
	return (rcode==0);
}

int SendMessage(int direction, int value)
{
	switch(direction)
	{
		case DECP:
			devmessage.FDECP = value;
			break;
		case RAP:
			devmessage.FRAP = value;
			break;
		case RAM:
			devmessage.FRAM = value;
			break;
		case DECM:
			devmessage.FDECM = value;
			break;
	}
	databuffer[0] = devmessage.whole_byte;
	return(ToDevice(databuffer, sizeof(databuffer)));
}

/* Callback function for the xx command buttons */
void commandCB(w,client_data,call_data)
	Widget w;
	int client_data;
	XmAnyCallbackStruct *call_data;
{
   
	XmString xms;
	int i, cmd_number, answer = 0;

	/* identify the button */
	cmd_number=client_data;

	switch (cmd_number) 
	{
		/* Connect / disconnect telescope */
		case CMD_CONNECT:
			if (device_link) 
			{
				if (CloseDevice())
				{
					device_link = 0;
					xms = XmStringCreateLocalized("Connect");
					XtVaSetValues (command[CMD_CONNECT], XmNlabelString, xms, NULL);
					XmStringFree(xms);
					setMessage("Device disconnected\n");
				}
			}
			else
			{
				if (OpenDevice(VENDOR_ID, PRODUCT_ID))
				{
					device_link = 1;
					xms = XmStringCreateLocalized("Disconnect");
					XtVaSetValues (command[CMD_CONNECT], XmNlabelString, xms, NULL);
					XmStringFree(xms);
					setMessage("Device connected\n");
					if (FromDevice(databuffer, sizeof(databuffer)))
					{
						devmessage.whole_byte = databuffer[0]; 
						sprintf(strmsg, "Status:\nRa- %u Ra+ %u\nDec- %u Dec+ %u\nLed color %u\nLed enable %u\nRes1 %u Res2 %u\n", devmessage.FRAP, devmessage.FRAM, devmessage.FDECP, devmessage.FDECM, devmessage.FLEDColor, devmessage.FLEDEnable, devmessage.FRes1, devmessage.FRes2);
						setMessage(strmsg);
					}
					else
					{
						fprintf(stderr, "Error communicating with device\n");
					}
				}
			}
			break;

		case CMD_LEDCOL:
			if (device_link) 
			{
				if (devmessage.FLEDColor == 0)
				{
					devmessage.FLEDColor = 1;
				}
				else
				{
					devmessage.FLEDColor = 0;
				}
				databuffer[0] = devmessage.whole_byte;
				if (ToDevice(databuffer, sizeof(databuffer)))
				{
					if (FromDevice(databuffer, sizeof(databuffer)))
					{
						devmessage.whole_byte = databuffer[0]; 
						sprintf(strmsg, "Led color %u\n", devmessage.FLEDColor);
						setMessage(strmsg);
					}
					else
					{
						fprintf(stderr, "Error communicating with device\n");
					}
				}
				else
				{
					fprintf(stderr, "Error communicating with device\n");
				}
			}
			else
			{
				setMessage("Device is not connected!\n");
			}
			break;

		case CMD_LEDENA:
			if (device_link) 
			{
				if (devmessage.FLEDEnable == 0)
				{
					devmessage.FLEDEnable = 1;
				}
				else
				{
					devmessage.FLEDEnable = 0;
				}
				databuffer[0] = devmessage.whole_byte;
				if (ToDevice(databuffer, sizeof(databuffer)))
				{
					if (FromDevice(databuffer, sizeof(databuffer)))
					{
						devmessage.whole_byte = databuffer[0]; 
						sprintf(strmsg, "Led enable %u\n", devmessage.FLEDEnable);
						setMessage(strmsg);
					}
					else
					{
						fprintf(stderr, "Error communicating with device\n");
					}
				}
				else
				{
					fprintf(stderr, "Error communicating with device\n");
				}
			}
			else
			{
				setMessage("Device is not connected!\n");
			}
			break;
	}
}

/* Callback function for the 4 directions buttons. */
/* Called when one of the grid buttons is pressed. */
void buttonCB(Widget w, XtPointer client_data, XtPointer call_data)
{
	int n;
	int errflag = FALSE;
	XmArrowButtonCallbackStruct *cbs = (XmArrowButtonCallbackStruct *) call_data;
	XtVaGetValues (w, XmNuserData, &n, NULL);

	if (device_link) 
	{  
		/* respond to the request */
		if (cbs->reason == XmCR_ARM)
		{
		
			switch (n) 
			{
				/* North */
				case XmARROW_UP:
					errflag = !SendMessage(DECP, 1);
					break;

				/* East */
				case XmARROW_RIGHT: 
					errflag = !SendMessage(RAM, 1);
					break;

				/* West */
				case XmARROW_LEFT: 
					errflag = !SendMessage(RAP, 1);
					break;

				/* South */
				case XmARROW_DOWN:
					errflag = !SendMessage(DECM, 1);
					break;
			}
		}

		if (cbs->reason == XmCR_DISARM)
		{
			switch (n) 
			{
				/* North */
				case XmARROW_UP:
					errflag = !SendMessage(DECP, 0);
					break;

				/* East */
				case XmARROW_RIGHT: 
					errflag = !SendMessage(RAM, 0);
					break;

				/* West */
				case XmARROW_LEFT: 
					errflag = !SendMessage(RAP, 0);
					break;

				/* South */
				case XmARROW_DOWN:
					errflag = !SendMessage(DECM, 0);
					break;
			}
		}
		switch (n) 
		{
			/* North */
			case XmARROW_UP:
				if (FromDevice(databuffer, sizeof(databuffer)))
				{
					sprintf(strmsg, "Dec+ %u\n", devmessage.FDECP);
					setMessage(strmsg);
				}
				break;

			/* East */
			case XmARROW_RIGHT: 
				if (FromDevice(databuffer, sizeof(databuffer)))
				{
					sprintf(strmsg, "Ra- %u\n", devmessage.FRAM);
					setMessage(strmsg);
				}
				break;

			/* West */
			case XmARROW_LEFT: 
				if (FromDevice(databuffer, sizeof(databuffer)))
				{
					sprintf(strmsg, "Ra+ %u\n", devmessage.FRAP);
					setMessage(strmsg);
				}
				break;

			/* South */
			case XmARROW_DOWN:
				if (FromDevice(databuffer, sizeof(databuffer)))
				{
					sprintf(strmsg, "Dec- %u\n", devmessage.FDECM);
					setMessage(strmsg);
				}
				break;
		}
		if (errflag)
		{
			setMessage("Error communicating with device\n"); 
		}
	} 
	else 
	{
		setMessage("Device is not connected\n"); 
	}
}

/* Callback function for the 2 swap options */
void swapCB(w, client_data, call_data)
	Widget w;
	int client_data;
	XmToggleButtonCallbackStruct *call_data;
{
	XmString text;

	if (call_data->reason == XmCR_VALUE_CHANGED)
	{
		/* identify the button code*/
		switch (client_data)
		{
			case 1:
			if (call_data->set == XmSET)
			{
				text = XmStringCreateLocalized ("W/E");
				XtVaSetValues (w, XmNlabelString, text, NULL);	
				XmStringFree(text);
				XtVaSetValues (button[CMDWEST], XmNuserData, XmARROW_LEFT, NULL);	
				XtVaSetValues (button[CMDEAST], XmNuserData, XmARROW_RIGHT, NULL);					
			}
			else
			{
				text = XmStringCreateLocalized ("E/W");
				XtVaSetValues (w, XmNlabelString, text, NULL);	
				XmStringFree(text);
				XtVaSetValues (button[CMDWEST], XmNuserData, XmARROW_RIGHT, NULL);	
				XtVaSetValues (button[CMDEAST], XmNuserData, XmARROW_LEFT, NULL);					
			}
			break;

			case 0:
			if (call_data->set == XmSET)
			{
				text = XmStringCreateLocalized ("S/N");
				XtVaSetValues (w, XmNlabelString, text, NULL);	
				XmStringFree(text);
				XtVaSetValues (button[CMDNORTH], XmNuserData, XmARROW_DOWN, NULL);	
				XtVaSetValues (button[CMDSOUTH], XmNuserData, XmARROW_UP, NULL);					
			}
			else
			{
				text = XmStringCreateLocalized ("N/S");
				XtVaSetValues (w, XmNlabelString, text, NULL);	
				XmStringFree(text);
				XtVaSetValues (button[CMDNORTH], XmNuserData, XmARROW_UP, NULL);	
				XtVaSetValues (button[CMDSOUTH], XmNuserData, XmARROW_DOWN, NULL);					
			}
			break;
		}
	}
}

/* Create Main Window */
void create_main_window()
{
	Arg al[15];
	int ac;
	int x;
	
	// Note that locations of widgts are in pixels

	/* Set the default size of the window */ 
	ac=0;
	XtSetArg(al[ac],XmNwidth,235); ac++;
	XtSetArg(al[ac],XmNheight,360); ac++;
	XtSetArg(al[ac],XmNminWidth,235); ac++;
	XtSetArg(al[ac],XmNminHeight,360); ac++;
	XtSetValues(toplevel,al,ac);

	/* Create a form widget */
	ac=0;
	form=XmCreateForm(toplevel,"form",al,ac);
	XtManageChild(form);

	/* Set up the command buttons and attach them to the form */
	for (x=0; x<CMD_BUTTONS; x++)   
	{
		ac=0;
		XtSetArg(al[ac],XmNlabelString, XmStringCreateLocalized(cmd_name[x])); ac++;
		XtSetArg(al[ac],XmNrecomputeSize, FALSE); ac++;
		XtSetArg(al[ac],XmNx, cmd_pos[x][0]); ac++;
		XtSetArg(al[ac],XmNy, cmd_pos[x][1]); ac++;
		XtSetArg(al[ac],XmNwidth, cmd_pos[x][2]); ac++;
		XtSetArg(al[ac],XmNheight, cmd_pos[x][3]); ac++;
		command[x]=XmCreatePushButton(form,"label",al,ac);
		XtManageChild(command[x]);
		XtAddCallback(command[x],XmNactivateCallback, commandCB,(XtPointer) x);
	}

	/* Set up the slew buttons, add event handlers, attach to form.*/	
	for (x=0; x<4; x++)
	{
		ac=0;
		XtSetArg(al[ac],XmNuserData, btn_name[x]); ac++;
		XtSetArg(al[ac],XmNy, btn_pos[x][1]); ac++;
		XtSetArg(al[ac],XmNx, btn_pos[x][0]); ac++;
		XtSetArg(al[ac],XmNwidth, 65); ac++;
		XtSetArg(al[ac],XmNheight, 65); ac++;
		XtSetArg(al[ac],XmNarrowDirection, btn_name[x]); ac++;
		XtSetArg(al[ac], XmNframeChildType, XmFRAME_GENERIC_CHILD); ac++;
		button[x]=XmCreateArrowButtonGadget(form,"label",al,ac);
		XtAddCallback (button[x], XmNarmCallback, buttonCB, (XtPointer) x);
		XtAddCallback (button[x], XmNdisarmCallback, buttonCB, (XtPointer) x);         
		XtManageChild(button[x]);
	}

	// Direction arrow reverse checks
	ac=0;
	XtSetArg(al[ac],XmNlabelString, XmStringCreateLocalized("N/S")); ac++;
	XtSetArg(al[ac],XmNrecomputeSize, FALSE); ac++;
	XtSetArg(al[ac],XmNy, 35); ac++;
	XtSetArg(al[ac],XmNx, 14); ac++;
	XtSetArg(al[ac],XmNwidth, 50); ac++;
	XtSetArg(al[ac],XmNheight, 21); ac++;
	XtSetArg(al[ac],XmNalignment,XmALIGNMENT_BEGINNING); ac++;
	XtSetArg(al[ac],XmNborderWidth,1); ac++;
	XtSetArg(al[ac],XmNborderColor,0); ac++;
	XtSetArg(al[ac], XmNframeChildType, XmFRAME_GENERIC_CHILD); ac++;
	toggleS[0]=XmCreateToggleButton(form,"toggle",al,ac);
	XtManageChild(toggleS[0]);
	XtAddCallback (toggleS[0], XmNvalueChangedCallback, swapCB, (XtPointer) 0);

	ac=0;
	XtSetArg(al[ac],XmNlabelString, XmStringCreateLocalized("E/W")); ac++;
	XtSetArg(al[ac],XmNrecomputeSize, FALSE); ac++;
	XtSetArg(al[ac],XmNy,57); ac++;
	XtSetArg(al[ac],XmNx, 14); ac++;
	XtSetArg(al[ac],XmNwidth, 50); ac++;
	XtSetArg(al[ac],XmNheight, 21); ac++;
	XtSetArg(al[ac],XmNalignment,XmALIGNMENT_BEGINNING); ac++;
	XtSetArg(al[ac],XmNborderWidth,1); ac++;
	XtSetArg(al[ac],XmNborderColor,0); ac++;
	XtSetArg(al[ac], XmNframeChildType, XmFRAME_GENERIC_CHILD); ac++;
	toggleS[1]=XmCreateToggleButton(form,"toggle",al,ac);
	XtManageChild(toggleS[1]);
	XtAddCallback (toggleS[1], XmNvalueChangedCallback, swapCB, (XtPointer) 1);

	/* message area*/
	/* Create a message widget and attach it to the form */
	ac=0;
	XtSetArg(al[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
	XtSetArg(al[ac], XmNtopOffset, 230); ac++;
	XtSetArg(al[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
	XtSetArg(al[ac], XmNleftOffset, 12); ac++;
	XtSetArg(al[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
	XtSetArg(al[ac], XmNrightOffset, 12); ac++;
	XtSetArg(al[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
	XtSetArg(al[ac], XmNbottomOffset, 12); ac++;
	XtSetArg(al[ac], XmNeditable,False); ac++;
	XtSetArg(al[ac], XmNeditMode,XmMULTI_LINE_EDIT); ac++;
	XtSetArg(al[ac], XmNwordWrap, True); ac++;
	message_area=XmCreateScrolledText(form,"message_area",al,ac);
	XtManageChild(message_area);

}

void close_app_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
	if (device_link) 
	{
		CloseDevice();
	}
	
	exit(0);
	return;
}

int main(argc,argv)
	int argc; 
	char *argv[];
{
	/* Create the toplevel shell */
	toplevel = XtAppInitialize(&context,"",NULL,0,&argc,argv, NULL,NULL,0);

	/* Create Main Window */
	create_main_window();  

	/* Start the user interface */
	XtRealizeWidget(toplevel);

	// The application close callback to manage the program end gracefully
	Atom wm_delete_window = XmInternAtom(XtDisplay(toplevel), "WM_DELETE_WINDOW", FALSE);
	XmAddWMProtocolCallback(toplevel, wm_delete_window, close_app_callback, NULL); 

	XtAppMainLoop(context);

	exit(0);
}

